; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Files manipulation library. KolibriOS part.
;
;  Target OS: KolibriOS
;
;  Dependencies: Uses system calls for file processing and memory.asm for memory
;                allocations.
;
;  Notes:
;_________________________________________________________________________________________

DIR_SLASH equ '/'

fsFromBegin     = 0
fsFromEnd       = 2
fsFromCurrent   = 1

accessRead  = 1
accessWrite = 2


fileErrorUndefinedBase = 1
fileErrorFunctionNotAvailable = 2
fileErrorUnknownFileSystem    = 3
fileErrorFileNotFound         = 5
fileErrorEOF                  = 6
fileErrorBadPointer           = 7
fileErrorDiskFull             = 8
fileErrorBrokenFAT            = 9
fileErrorAccessDenied         = 10
fileErrorDeviceError          = 11
fileErrorOutOfMemory          = 30
fileErrorNotExecutable        = 31
fileErrorTooManyProcesses     = 32


if used STDIN | used STDOUT | used STDERR

initialize InitStdHandles
begin
  if used STDIN | used STDOUT
        mov     eax, [STDIO]
  end if

  if used STDIN
        mov     dword [__std_handle_in], eax
  end if

  if used STDOUT
        mov     dword [__std_handle_out], eax
  end if

  if used STDERR
        mov     eax, [STDDBG]
        mov     dword [__std_handle_err], eax
  end if
        return
endp

end if


faReadOnly = 1
faHidden   = 2
faSystem   = 4
faVolume   = 8
faDirectory= 16
faArchive  = 32


struct __TFileInfo
  .attr          dd ?
  .encoding      db ?
                 rb 3
  .creation_time dd ?
  .creation_date dd ?
  .access_time   dd ?
  .access_date   dd ?
  .modify_time   dd ?
  .modify_date   dd ?
  .file_size     dd ?
  .file_sizeH    dd ?
ends


struct __TFileOp
  .subfunc   dd ?
  .offset    dd ?
  .flags     dd ?
  .size      dd ?
  .ptrBuffer dd ?
  .filename    db 0
  .ptrFilename dd ?
ends


struct __TFile
  .hFilename dd ?
  .pointer   dd ?
  .access    dd ?

  .info __TFileInfo
ends


;--------------------------------------
; if CF = 0, eax is handle to the file.
; if CF = 1, eax is error code.
;--------------------------------------
proc FileOpen, .filename
.op   __TFileOp
begin
        push    ebx esi

        stdcall GetMem, sizeof.__TFile
        mov     esi, eax

        stdcall StrDup, [.filename]
        mov     [esi+__TFile.hFilename], eax
        mov     [esi+__TFile.pointer], 0
        mov     [esi+__TFile.access], accessRead

        lea     ebx, [.op]
        xor     eax, eax

        mov     [ebx+__TFileOp.subfunc], 5
        mov     [ebx+__TFileOp.offset], eax
        mov     [ebx+__TFileOp.flags], eax
        mov     [ebx+__TFileOp.size], eax
        mov     [ebx+__TFileOp.filename], al
        lea     eax, [esi+__TFile.info]
        mov     [ebx+__TFileOp.ptrBuffer], eax
        stdcall StrPtr, [esi+__TFile.hFilename]
        mov     [ebx+__TFileOp.ptrFilename], eax

        mov     eax, sys_file_system_lfn
        int     $40

        test    eax, eax
        jnz     .error

        mov     eax, esi
        clc
        pop     esi ebx
        return

.error:
        stdcall FileClose, esi
        stc
        pop     esi ebx
        return
endp



;--------------------------------------
; if CF = 0, eax is handle to the file.
; if CF = 1, eax is error code.
;--------------------------------------
proc FileCreate, .filename
.op   __TFileOp
.buff dd ?
begin
        push    ebx esi

        stdcall GetMem, sizeof.__TFile
        mov     esi, eax

        stdcall StrDup, [.filename]
        mov     [esi+__TFile.hFilename], eax
        mov     [esi+__TFile.pointer], 0
        mov     [esi+__TFile.access], accessWrite

        lea     ebx, [.op]
        xor     eax, eax

        mov     [ebx+__TFileOp.subfunc], 2
        mov     [ebx+__TFileOp.offset], eax
        mov     [ebx+__TFileOp.flags], eax
        mov     [ebx+__TFileOp.size], eax
        mov     [ebx+__TFileOp.filename], al
        lea     eax, [esi+__TFile.info]
        mov     [ebx+__TFileOp.ptrBuffer], eax
        stdcall StrPtr, [esi+__TFile.hFilename]
        mov     [ebx+__TFileOp.ptrFilename], eax

        mov     eax, sys_file_system_lfn
        int     $40
        test    eax, eax
        jnz     .error

        lea     ebx, [.op]
        mov     [ebx+__TFileOp.subfunc], 5
        int     $40

        test    eax, eax
        jnz     .error

        mov     eax, esi
        clc
        pop     esi ebx
        return

.error:
        stdcall FileClose, esi
        stc
        pop     esi ebx
        return
endp




;--------------------------------------
; if CF = 0, the file was closed.
; if CF = 1, eax is error code.
;--------------------------------------
proc FileClose, .handle
begin
        push    eax
        mov     eax, [.handle]
        stdcall StrDel, [eax+__TFile.hFilename]
        stdcall FreeMem, eax
        clc
        pop     eax
        return
endp


; returns 32bit file size in eax
proc FileSize, .handle
begin
        mov     eax, [.handle]
        mov     eax, [eax+__TFile.info.file_size]
        clc
        return
endp



proc FileRead, .handle, .buffer, .count
.op   __TFileOp
begin
        push    ebx esi

        mov     esi, [.handle]

        cmp     esi, [STDIN]
        je      .console_read

        cmp     esi, [STDOUT]
        je      .console_read

        cmp     esi, [STDERR]
        je      .console_read

        lea     ebx, [.op]
        xor     eax, eax

        mov     [ebx+__TFileOp.subfunc], eax
        mov     [ebx+__TFileOp.flags], eax
        mov     [ebx+__TFileOp.filename], al

        mov     eax, [esi+__TFile.pointer]
        mov     [ebx+__TFileOp.offset], eax

        mov     eax, [.count]
        mov     [ebx+__TFileOp.size], eax

        mov     eax, [.buffer]
        mov     [ebx+__TFileOp.ptrBuffer], eax

        stdcall StrPtr, [esi+__TFile.hFilename]
        mov     [ebx+__TFileOp.ptrFilename], eax

        mov     eax, sys_file_system_lfn
        int     $40

        test    eax, eax
        jz      .readok

        cmp     eax, fileErrorEOF
        jne     .error

.readok:
        add     [esi+__TFile.pointer], ebx
        mov     eax, ebx
        clc
        pop     esi ebx
        return

.error:
        stc
        pop     esi ebx
        return

.console_read:
        mov     esi, [.buffer]

.read_loop:
        stdcall SocketReceive, [.handle], esi, 1, 0
        jc      .finish

        inc     esi

        cmp     byte [esi-1], $0d
        je      .return

        dec     [.count]
        jnz     .read_loop

.return:
        sub     esi, [.buffer]
        mov     eax, esi

.finish:
        pop     esi ebx
        return
endp




proc FileWrite, .handle, .buffer, .count
.op   __TFileOp
begin
        push    ebx esi

        mov     esi, [.handle]

        cmp     esi, [STDIN]
        je      .console_write

        cmp     esi, [STDOUT]
        je      .console_write

        cmp     esi, [STDERR]
        je      .console_write


        lea     ebx, [.op]
        xor     eax, eax

        mov     [ebx+__TFileOp.subfunc], 3
        mov     [ebx+__TFileOp.flags], eax
        mov     [ebx+__TFileOp.filename], al

        mov     eax, [esi+__TFile.pointer]
        mov     [ebx+__TFileOp.offset], eax

        mov     eax, [.count]
        mov     [ebx+__TFileOp.size], eax

        mov     eax, [.buffer]
        mov     [ebx+__TFileOp.ptrBuffer], eax

        stdcall StrPtr, [esi+__TFile.hFilename]
        mov     [ebx+__TFileOp.ptrFilename], eax

        mov     eax, sys_file_system_lfn
        int     $40

        add     ebx, [esi+__TFile.pointer]
        cmp     [esi+__TFile.info.file_size], ebx
        ja      @f
        mov     [esi+__TFile.info.file_size], ebx
@@:
        mov     [esi+__TFile.pointer], ebx

        cmp     eax, 1
        cmc                     ; CF = 1 if eax==0;

        mov     eax, ebx
        pop     esi ebx
        return


.console_write:

        stdcall SocketSend, [.handle], [.buffer], [.count], 0
        pop     esi ebx
        return

endp


;----------------------------------------------------
; if CF = 0, eax is the new possition in the file
; if CF = 1, eax is error code.
;----------------------------------------------------
proc FileSeek, .handle, .dist, .direction
begin
        push    esi
        mov     esi, [.handle]
        mov     eax, [.dist]

        cmp     [.direction], fsFromCurrent
        je      .rel
        cmp     [.direction], fsFromBegin
        je      .begin
        cmp     [.direction], fsFromEnd
        je      .end

.error:
        stc
        mov     eax, [esi+__TFile.pointer]
        pop     esi
        return

.ok:
        clc
        pop     esi
        return

.rel:
        add     eax, [esi+__TFile.pointer]

.begin:
        cmp     [esi+__TFile.info.file_size], eax
        jae     @f
        mov     eax, [esi+__TFile.info.file_size]
@@:
        mov     [esi+__TFile.pointer], eax
        jmp     .ok

.end:
        add     eax, [esi+__TFile.info.file_size]
        js      .error
        jmp     .begin
endp



;--------------------------------------
; if CF = 0, the file was deleted.
; if CF = 1, eax is error code.
;--------------------------------------
proc FileDelete, .filename
.op   __TFileOp
begin
        push    ebx

        lea     ebx, [.op]
        xor     eax, eax

        mov     [ebx+__TFileOp.offset], eax
        mov     [ebx+__TFileOp.size], eax
        mov     [ebx+__TFileOp.ptrBuffer], eax
        mov     [ebx+__TFileOp.flags], eax
        mov     [ebx+__TFileOp.filename], al

        stdcall StrPtr, [esi+__TFile.hFilename]
        mov     [ebx+__TFileOp.ptrFilename], eax

        mov     eax, sys_file_system_lfn
        mov     [ebx+__TFileOp.subfunc], 8
        int     $40

        cmp     eax, 1
        cmc
        pop     ebx
        return
endp





proc GetCurrentDir
.buffer rb 1024
begin
        push    ebx ecx edx

        mov     eax, sys_current_directory
        mov     ebx, 2          ; get current dir
        lea     ecx, [.buffer]
        mov     edx, 1024
        int     $40

        stdcall StrDup, ecx
        push    eax

        stdcall StrPtr, eax
        mov     ecx, [eax+string.len]
        cmp     ecx, 1
        je      @f

        cmp     byte [eax+ecx-1], '/'
        jne     @f
        dec     [eax+string.len]
        mov     dword [eax+ecx-1], 0
@@:
        pop     eax
        clc
        pop     ecx ebx
        return
endp




proc SetCurrentDir, .hDirectory
begin
        push    eax ebx ecx

        stdcall StrPtr, [.hDirectory]
        mov     ecx, eax
        mov     eax, sys_current_directory
        mov     ebx, 1  ; set current dir
        int     $40

        clc
        pop     ecx ebx eax
        return
endp







proc GetErrorString, .code
begin
        mov     eax, [.code]
        test    eax, eax
        jz      .finish

        cmp     eax, 30
        jb      .normal

        sub     eax, 30
        cmp     eax, 2
        ja      .unknown

        mov     eax, [.error2+4*eax]
        jmp     .finish

.normal:
        cmp     eax, 11
        ja      .unknown

        mov     eax, [.error+4*eax]

.finish:
        clc
        return

.unknown:
        mov     eax, cErrorUnknown
        jmp     .finish

.error   dd cErrorOK, cErrorMissingBase, cErrorFnNotSupported, cErrorUnknownFS, cErrorReserved, cErrorFileNotFound, cErrorEndOfFile, cErrorBadPointer, cErrorDiskFull, cErrorBadFAT, cErrorAccessDenied, cErrorDeviceError
.error2  dd cErrorOutOfMemory, cErrorNotExecutable, cErrorTooManyProcesses
endp

cErrorOK             text    'No error.'
cErrorMissingBase    text    'Missing base.'      ;???
cErrorFnNotSupported text    'Function not supported on this FS.'
cErrorUnknownFS      text    'Not supported file system.'
cErrorReserved       text    'Reserved error code! Kernel bug?'
cErrorFileNotFound   text    'File not found.'
cErrorEndOfFile      text    'End of file reached.'
cErrorBadPointer     text    'Bad buffer pointer.'
cErrorDiskFull       text    'Disk full.'
cErrorBadFAT         text    'Bad FAT.'
cErrorAccessDenied   text    'Access denied.'
cErrorDeviceError    text    'Device error.'

cErrorUnknown        text    'Unknown error code.'

cErrorOutOfMemory    text    'Out of memory.'
cErrorNotExecutable  text    'File is not executable.'
cErrorTooManyProcesses text  'Too many processes.'



proc FreeErrorString, .ptrString
begin
        return
endp




ftDirectory  =  1
ftUnknown    =  2
ftFile       =  3


; reads all items of the directory and fills them in the TArray of type TDirItem.
; arguments:
;   dir - string with the path to the directory.
;
; returns:
;   if not error: CF=0; eax = pointer to TArray of TDirItem;
;   if error:     CF=1; eax = error code;

proc DirectoryRead, .dir
.op       __TFileOp
          rb  256
.buffer   rb  1024
begin
        push    ebx ecx edx esi edi

        stdcall StrDup, [.dir]
        mov     [.dir], eax

        xor     esi, esi        ; file number

        stdcall CreateArray, sizeof.TDirItem
        mov     edx, eax

.loop:
        lea     ebx, [.op]
        mov     [ebx+__TFileOp.subfunc], fsfReadDir
        mov     [ebx+__TFileOp.offset], esi
        mov     [ebx+__TFileOp.flags], 0        ; ASCII
        mov     [ebx+__TFileOp.size], 1         ; read one block
        lea     edi, [.buffer]
        mov     [ebx+__TFileOp.ptrBuffer], edi

        xor     eax, eax
        mov     ecx, 1024/4
        rep stosd

        mov     [ebx+__TFileOp.filename], 0
        stdcall StrPtr, [.dir]
        mov     [ebx+__TFileOp.ptrFilename], eax

        mov     eax, sys_file_system_lfn
        int     $40

        test    eax, eax
        jnz     .end_read

        test    dword [.buffer+$20], $08      ; flag volume label
        jnz     .next

        lea     eax, [.buffer+$48]      ; filename in widechar (UTF-16)
        stdcall StrDup, eax
;        stdcall WideCharToUtf8, eax
        mov     ecx, eax

        stdcall AddArrayItems, edx, 1

        mov     [eax+TDirItem.hFilename], ecx

        mov     ecx, ftDirectory
        test    dword [.buffer+$20], $10      ; flag directory
        jnz     @f
        mov     ecx, ftFile
@@:
        mov     [eax+TDirItem.Type], ecx

.next:
        inc     esi
        jmp     .loop

.end_read:
        stdcall StrDel, [.dir]

        cmp     esi, dword [.buffer+8]       ; the count of the elements in the directory.
        jae     .finish_ok

        stdcall FreeDirArray, edx
        xor     eax, eax
        stc
        pop     esi edx ecx ebx
        return


.finish_ok:
        mov     eax, edx        ; the TArray
        clc
        pop     edi esi edx ecx ebx
        return

endp





proc GetFileInfo, .hFile, .ptrFileInfo
begin
        stc
        return
endp



proc __ConsoleWrite, .ptrData, .count
begin
        push    eax ecx esi

        mov     esi, [.ptrData]
        mov     ecx, [.count]

.loop:
        test    ecx, ecx
        jz      .end_output

        dec     ecx
        movzx   eax, byte [esi]
        lea     esi, [esi+1]

        stdcall CharOut, eax
        jmp     .loop

.end_output:
        pop     esi ecx eax
        return
endp